# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import os
import yaml

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils
from toscaparser import tosca_template

from tacker.common import exceptions
from tacker.common import utils
from tacker.extensions import nfvo
from tacker import objects
from tacker.objects import fields
from tacker.tosca import utils as toscautils
from tacker.vnfm import vim_client

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


def _get_vim(context, vim_connection_info):
    vim_client_obj = vim_client.VimClient()

    if vim_connection_info:
        vim_id = vim_connection_info[0].vim_id
        access_info = vim_connection_info[0].access_info
        if access_info:
            region_name = access_info.get('region')
        else:
            region_name = None
    else:
        vim_id = None
        region_name = None

    try:
        vim_res = vim_client_obj.get_vim(
            context, vim_id, region_name=region_name)
    except nfvo.VimNotFoundException:
        raise exceptions.VimConnectionNotFound(vim_id=vim_id)

    vim_res['vim_auth'].update({'region': region_name})
    vim_info = {'id': vim_res['vim_id'], 'vim_id': vim_res['vim_id'],
                'vim_type': vim_res['vim_type'],
                'access_info': vim_res['vim_auth']}

    return vim_info


def _get_vnfd_dict(context, vnfd_id, flavour_id):
    vnfd_dict = _get_flavour_based_vnfd(
        _get_vnf_package_path(context, vnfd_id), flavour_id)

    # Remove requirements from substitution mapping
    vnfd_dict.get('topology_template').get(
        'substitution_mappings').pop('requirements')
    return vnfd_dict


def _get_vnflcm_interface(context, interface, vnf_instance, flavour_id):
    '''Gets the interface found in vnfd

        ...
            node_templates:
                VNF:
                    interfaces:
                        Vnflcm:
                            <interface>
    '''
    interface_value = None
    vnfd_dict = _get_vnfd_dict(context, vnf_instance.vnfd_id, flavour_id)

    if not isinstance(vnfd_dict, dict):
        raise exceptions.InvalidContentType(msg="VNFD not valid")

    if vnfd_dict.get('topology_template'):
        topology_template = vnfd_dict.get('topology_template')
        if topology_template.get('node_templates'):
            node_templates = topology_template.get('node_templates')
            if node_templates.get('VNF'):
                vnf = node_templates.get('VNF')
                if vnf.get('interfaces'):
                    interfaces = vnf.get('interfaces')
                    if interfaces.get('Vnflcm'):
                        vnflcm = interfaces.get('Vnflcm')
                        if vnflcm:
                            interface_value = vnflcm.get(interface)

    return interface_value


def _build_affected_resources(vnf_instance,
     change_type=fields.ResourceChangeType.ADDED):
    '''build affected resources from vnf_instance instantiated info '''

    affected_resources = {}
    instantiated_vnf_info = vnf_instance.instantiated_vnf_info
    if hasattr(instantiated_vnf_info, 'instance_id'):
        if instantiated_vnf_info.instance_id:

            affected_resources['affectedVnfcs'] = []
            affected_resources['affectedVirtualLinks'] = []
            affected_resources['affectedVirtualStorages'] = []

            # build AffectedVnfc
            vnfc_resource_info = \
                instantiated_vnf_info.vnfc_resource_info
            for vnfc_resource in vnfc_resource_info:
                data = {}
                data['id'] = vnfc_resource.id
                data['vduId'] = vnfc_resource.vdu_id
                data['changeType'] = change_type
                data['computeResource'] = \
                    vnfc_resource.compute_resource.to_dict()
                data['metadata'] = vnfc_resource.metadata
                affected_resources['affectedVnfcs'].append(data)

            # build AffectedVirtualLink
            vnf_virtual_link = \
                instantiated_vnf_info.vnf_virtual_link_resource_info
            for vnf_vl_info in vnf_virtual_link:
                data = {}
                data['id'] = vnf_vl_info.id
                data['vnfVirtualLinkDescId'] = \
                    vnf_vl_info.vnf_virtual_link_desc_id
                data['changeType'] = change_type
                data['networkResource'] = \
                    vnf_vl_info.network_resource.to_dict()
                data['metadata'] = {}
                affected_resources['affectedVirtualLinks'].append(data)

            # build affectedVirtualStorages
            virtual_storage = \
                instantiated_vnf_info.virtual_storage_resource_info
            for vnf_storage_info in virtual_storage:
                data = {}
                data['id'] = vnf_storage_info.id
                data['virtualStorageDescId'] = \
                    vnf_storage_info.virtual_storage_desc_id
                data['changeType'] = change_type
                data['storageResource'] = \
                    vnf_storage_info.storage_resource.to_dict()
                data['metadata'] = {}
                affected_resources['affectedVirtualStorages'].append(data)

    return utils.convert_snakecase_to_camelcase(affected_resources)


def _get_affected_resources(old_vnf_instance=None,
                new_vnf_instance=None, extra_list=None):
    '''get_affected_resources

    returns affected resources in new_vnf_instance not present
    in old_vnf_instance.
    if extra_list (list of physical resource ids) is present,
        included to affected resources
    '''
    def _get_affected_cpids(affected_vnfc, vnf_instance):
        affected_cpids = []
        instantiated_vnf_info = vnf_instance.instantiated_vnf_info
        for vnfc_resource in instantiated_vnf_info.vnfc_resource_info:
            if vnfc_resource.id == affected_vnfc['id']:
                for vnfc_cp in vnfc_resource.vnfc_cp_info:
                    if vnfc_cp.cpd_id:
                        affected_cpids.append(vnfc_cp.cpd_id)
                    if vnfc_cp.vnf_ext_cp_id:
                        affected_cpids.append(vnfc_cp.vnf_ext_cp_id)
        return affected_cpids

    def _get_added_storageids(affected_vnfc, vnf_instance):
        affected_storage_ids = []
        instantiated_vnf_info = vnf_instance.instantiated_vnf_info
        for vnfc_resource in instantiated_vnf_info.vnfc_resource_info:
            if vnfc_resource.id == affected_vnfc['id']:
                for storage_resource_id in vnfc_resource.storage_resource_ids:
                    virtual_storage = \
                        instantiated_vnf_info.virtual_storage_resource_info
                    for virt_storage_res_info in virtual_storage:
                        if virt_storage_res_info.id == storage_resource_id:
                            affected_storage_ids.append(
                                virt_storage_res_info.virtual_storage_desc_id)
        return affected_storage_ids

    def diff_list(old_list, new_list):
        diff = []
        for item in new_list:
            if item not in old_list:
                diff.append(item)
        return diff

    affected_resources = {}
    affected_resources['affectedVnfcs'] = []
    affected_resources['affectedVirtualLinks'] = []
    affected_resources['affectedVirtualStorages'] = []

    if not old_vnf_instance:
        affected_resources = _build_affected_resources(
            new_vnf_instance, fields.ResourceChangeType.ADDED)
        # add affected cpids and add added storageids
        for affected_vnfc in affected_resources['affectedVnfcs']:
            affected_vnfc['affectedVnfcCpIds'] = _get_affected_cpids(
                affected_vnfc, new_vnf_instance)
            affected_vnfc['addedStorageResourceIds'] = _get_added_storageids(
                affected_vnfc, new_vnf_instance)

    elif not new_vnf_instance:
        affected_resources = _build_affected_resources(old_vnf_instance,
                                            fields.ResourceChangeType.REMOVED)
        # add affected cpids and add remove storageids
        for affected_vnfc in affected_resources['affectedVnfcs']:
            affected_vnfc['affectedVnfcCpIds'] = _get_affected_cpids(
                affected_vnfc, old_vnf_instance)
            affected_vnfc['removedStorageResourceIds'] = _get_added_storageids(
                affected_vnfc, old_vnf_instance)
    elif old_vnf_instance and new_vnf_instance:
        old_affected_resources = _build_affected_resources(old_vnf_instance)
        new_affected_resources = _build_affected_resources(new_vnf_instance,
                                            fields.ResourceChangeType.MODIFIED)

        # get resource_ids
        old_vnfc_resource_ids = []
        for vnfc_resource in old_affected_resources.get('affectedVnfcs', []):
            old_vnfc_resource_ids.append(
                vnfc_resource['computeResource']['resourceId'])

        # remove extra_list items in old_vnfc_resource_ids
        # so that this items will be considered new
        if extra_list:
            for item in extra_list:
                if item in old_vnfc_resource_ids:
                    index = old_vnfc_resource_ids.index(item)
                    old_vnfc_resource_ids.pop(index)

        new_vnfc_resource_ids = []
        for vnfc_resource in new_affected_resources.get('affectedVnfcs', []):
            resource_id = vnfc_resource['computeResource']['resourceId']
            new_vnfc_resource_ids.append(resource_id)

        old_vnf_vl_resource_ids = []
        for vnf_vl_info in old_affected_resources.get(
                'affectedVirtualLinks', []):
            resource_id = vnf_vl_info['networkResource']['resourceId']
            old_vnf_vl_resource_ids.append(resource_id)

        new_vnf_vl_resource_ids = []
        for vnf_vl_info in new_affected_resources.get(
                'affectedVirtualLinks', []):
            resource_id = vnf_vl_info['networkResource']['resourceId']
            new_vnf_vl_resource_ids.append(resource_id)

        old_vnf_storage_resource_ids = []
        for vnf_storage_info in old_affected_resources.get(
                'affectedVirtualStorages', []):
            resource_id = vnf_storage_info['storageResource']['resourceId']
            old_vnf_storage_resource_ids.append(resource_id)

        new_vnf_storage_resource_ids = []
        for vnf_storage_info in new_affected_resources.get(
                'affectedVirtualStorages', []):
            resource_id = vnf_storage_info['storageResource']['resourceId']
            new_vnf_storage_resource_ids.append(resource_id)

        # get difference between resource_ids
        vnfc_resource_ids = diff_list(old_vnfc_resource_ids,
                                      new_vnfc_resource_ids)
        vnf_vl_resource_ids = diff_list(old_vnf_vl_resource_ids,
                                        new_vnf_vl_resource_ids)
        vnf_storage_resource_ids = diff_list(old_vnf_storage_resource_ids,
                                             new_vnf_storage_resource_ids)

        # return new affected resources
        for affected_vls in new_affected_resources['affectedVirtualLinks']:
            if (affected_vls['networkResource']
                    ['resourceId'] in vnf_vl_resource_ids):
                affected_resources['affectedVirtualLinks'].append(affected_vls)

        affected_storages = new_affected_resources['affectedVirtualStorages']
        for affected_storage in affected_storages:
            if (affected_storage['storageResource']
                    ['resourceId'] in vnf_storage_resource_ids):
                affected_resources['affectedVirtualStorages'].append(
                    affected_storage)

        for affected_vnfc in new_affected_resources['affectedVnfcs']:
            if (affected_vnfc['computeResource']
                    ['resourceId'] in vnfc_resource_ids):

                # update affected affectedVnfcCpIds
                affected_vnfc['affectedVnfcCpIds'] = _get_affected_cpids(
                    affected_vnfc, new_vnf_instance)

                affected_resources['affectedVnfcs'].append(affected_vnfc)

    return affected_resources


def _get_vnf_package_id(context, vnfd_id):
    vnf_package = objects.VnfPackageVnfd.get_by_id(context, vnfd_id)
    return vnf_package.package_uuid


def _create_grant_request(vnfd_dict, package_uuid):
    node_templates = vnfd_dict.get('topology_template',
                                   {}).get('node_templates', {})
    vnf_software_images = {}
    if not node_templates:
        return vnf_software_images

    def _build_vnf_software_image(sw_image_data, artifact_image_path):
        vnf_sw_image = objects.VnfSoftwareImage()
        vnf_sw_image.image_path = artifact_image_path
        vnf_sw_image.name = sw_image_data.get('name')
        vnf_sw_image.version = sw_image_data.get('version')
        if sw_image_data.get('checksum'):
            checksum = sw_image_data.get('checksum')
            if checksum.get('algorithm'):
                vnf_sw_image.algorithm = checksum.get('algorithm')
            if checksum.get('hash'):
                vnf_sw_image.hash = checksum.get('hash')

        vnf_sw_image.container_format = sw_image_data.get(
            'container_format')
        vnf_sw_image.disk_format = sw_image_data.get('disk_format')
        if sw_image_data.get('min_disk'):
            min_disk = utils.MemoryUnit.convert_unit_size_to_num(
                sw_image_data.get('min_disk'), 'GB')
            vnf_sw_image.min_disk = min_disk
        else:
            vnf_sw_image.min_disk = 0

        if sw_image_data.get('min_ram'):
            min_ram = utils.MemoryUnit.convert_unit_size_to_num(
                sw_image_data.get('min_ram'), 'MB')
            vnf_sw_image.min_ram = min_ram
        else:
            vnf_sw_image.min_ram = 0

        return vnf_sw_image

    def _get_image_path(artifact_image_path, package_uuid):
        vnf_package_path = CONF.vnf_package.vnf_package_csar_path
        artifact_image_path = os.path.join(
            vnf_package_path, package_uuid,
            artifact_image_path.split('../')[-1])
        return artifact_image_path

    for node, value in node_templates.items():
        if not value.get(
                'type') in ['tosca.nodes.nfv.Vdu.Compute',
                            'tosca.nodes.nfv.Vdu.VirtualBlockStorage']:
            continue

        sw_image_data = value.get('properties', {}).get('sw_image_data')
        artifacts = value.get('artifacts', {})
        for artifact, sw_image in artifacts.items():
            artifact_image_path = None
            if isinstance(sw_image, str):
                artifact_image_path = sw_image
            elif sw_image.get('type') == 'tosca.artifacts.nfv.SwImage':
                artifact_image_path = sw_image.get('file', {})
            if sw_image_data and artifact_image_path:
                is_url = utils.is_url(artifact_image_path)
                if not is_url:
                    artifact_image_path = _get_image_path(artifact_image_path,
                                                          package_uuid)

                vnf_software_image = _build_vnf_software_image(
                    sw_image_data, artifact_image_path)
                vnf_software_images[node] = vnf_software_image
            break

    return vnf_software_images


def _make_final_vnf_dict(vnfd_dict, id, name, param_values, vnf_dict=None):
    if vnf_dict:
        final_vnf_dict = vnf_dict
        final_vnf_dict['vnfd']['attributes'].\
            update({'vnfd': str(vnfd_dict)})
        final_vnf_dict['attributes'].\
            update({'param_values': str(param_values)})
        final_vnf_dict['attributes'].\
            update({'stack_name': name or ("vnflcm_" + id)})
        return final_vnf_dict
    else:
        return {'vnfd': {
            'attributes': {
                'vnfd': str(vnfd_dict)}},
            'id': id,
            'name': name,
            'attributes': {
                'param_values': str(param_values),
                'stack_name': name or ("vnflcm_" + id)}}


def _get_flavour_based_vnfd(csar_path, flavour_id):
    ext = (".yaml", ".yml")
    file_path_and_data = {}
    imp_list = []
    for item in os.listdir(csar_path):
        src_path = os.path.join(csar_path, item)
        if os.path.isdir(src_path):
            for file in os.listdir(src_path):
                if file.endswith(ext):
                    source_file_path = os.path.join(src_path, file)
                    with open(source_file_path) as file_obj:
                        data = yaml.safe_load(file_obj)
                    substitution_map = data.get(
                        'topology_template',
                        {}).get('substitution_mappings', {})
                    if substitution_map.get(
                            'properties', {}).get('flavour_id') == flavour_id:
                        if data.get('imports'):
                            for imp in data.get('imports'):
                                imp_path = os.path.join(src_path, imp)
                                imp_list.append(imp_path)
                            data.update({'imports': imp_list})

                        return data

        elif src_path.endswith(ext):
            with open(src_path) as file_obj:
                file_data = yaml.safe_load(file_obj)
            substitution_map = file_data.get(
                'topology_template', {}).get('substitution_mappings', {})
            if substitution_map.get(
                    'properties', {}).get('flavour_id') == flavour_id:
                if file_data.get('imports'):
                    for imp in file_data.get('imports'):
                        imp_list.append(os.path.join(src_path, imp))
                    file_data.update({'imports': imp_list})
                return file_data

    return file_path_and_data


def _get_param_data(vnfd_dict, instantiate_vnf_req):
    param_value = {}
    additional_param = instantiate_vnf_req.additional_params
    if additional_param is None:
        additional_param = {}
    substitution_map = vnfd_dict.get('topology_template',
                                     {}).get('substitution_mappings', {})
    input_attributes = vnfd_dict.get('topology_template', {}).get('inputs')
    if substitution_map is not None:
        subs_map_node_type = substitution_map.get('node_type')
        # Get properties in lower-level VNFD for top-level VNFD
        node_templates = vnfd_dict.get('topology_template',
                                    {}).get('node_templates', {})
        for node in node_templates.values():
            if node.get('type') == subs_map_node_type:
                node_property = node.get('properties', {})
                if node_property:
                    param_value.update(node_property)
        # Import `_type.yaml` file and get default properties.
        # If new value provided in additional_param, the property is updated.
        import_paths = vnfd_dict.get('imports', {})
        for imp_path in import_paths:
            with open(imp_path) as file_obj:
                import_data = yaml.safe_load(file_obj)
            imp_node_type = import_data.get('node_types')
            if imp_node_type:
                for key, value in imp_node_type.items():
                    if key == subs_map_node_type:
                        properties = value.get('properties')
                        if properties:
                            for key, prop in properties.items():
                                if additional_param.get(key):
                                    param_value.update({
                                        key: additional_param.get(key)})
                                # If the parameter is provided in lower-level
                                # VNFD, use it. Otherwise use the default.
                                elif not param_value.get(key):
                                    param_value.update(
                                        {key: prop.get('default')})

    for input_attr, value in input_attributes.items():
        if additional_param.get(input_attr):
            param_value.update({input_attr: additional_param.get(
                input_attr)})

    return param_value


def _get_vim_connection_info_from_vnf_req(vnf_instance, instantiate_vnf_req):
    vim_connection_obj_list = []

    if not instantiate_vnf_req.vim_connection_info:
        # add default vim
        if len(vnf_instance.vim_connection_info):
            vim_connection_obj_list.append(vnf_instance.vim_connection_info[0])

        return vim_connection_obj_list

    for vim_connection in instantiate_vnf_req.vim_connection_info:
        vim_conn = objects.VimConnectionInfo(id=vim_connection.id,
            vim_id=vim_connection.vim_id, vim_type=vim_connection.vim_type,
            access_info=vim_connection.access_info)

        vim_connection_obj_list.append(vim_conn)

    # add default vim
    if len(vnf_instance.vim_connection_info):
        if vim_conn.id and vnf_instance.vim_connection_info[0].id:
            is_default_vim_exist = [vim_conn for vim_conn
                in vim_connection_obj_list
                    if vim_conn.id == vnf_instance.vim_connection_info[0].id]
            if not len(is_default_vim_exist):
                vim_connection_obj_list.append(vnf_instance.
                    vim_connection_info[0])

    return vim_connection_obj_list


def _build_instantiated_vnf_info(vnfd_dict, instantiate_vnf_req,
                                 vnf_instance, vim_id):
    inst_vnf_info = vnf_instance.instantiated_vnf_info
    inst_vnf_info.vnf_state = fields.VnfOperationalStateType.STARTED

    node_templates = vnfd_dict.get(
        'topology_template', {}).get('node_templates')

    vnfc_resource_info, virtual_storage_resource_info = \
        _get_vnfc_resource_info(vnfd_dict, instantiate_vnf_req, vim_id)

    inst_vnf_info.vnfc_resource_info = vnfc_resource_info

    tmp_insta_vnf_info = copy.deepcopy(inst_vnf_info)
    inst_vnf_info.ext_cp_info = _set_ext_cp_info(instantiate_vnf_req,
        inst_vnf_info=tmp_insta_vnf_info)
    inst_vnf_info.ext_virtual_link_info = _set_ext_virtual_link_info(
        instantiate_vnf_req, inst_vnf_info.ext_cp_info)

    inst_vnf_info.virtual_storage_resource_info = \
        virtual_storage_resource_info
    inst_vnf_info.vnf_virtual_link_resource_info = \
        _build_vnf_virtual_link_resource_info(
            node_templates, instantiate_vnf_req,
            inst_vnf_info.vnfc_resource_info, vim_id)

    inst_vnf_info.ext_managed_virtual_link_info = \
        _build_ext_managed_virtual_link_info(instantiate_vnf_req,
            inst_vnf_info)

    inst_vnf_info.additional_params = instantiate_vnf_req.additional_params

    vnf_instance.instantiated_vnf_info = inst_vnf_info


def _update_instantiated_vnf_info(change_ext_conn_req, vnf_instance):
    inst_vnf_info = vnf_instance.instantiated_vnf_info
    tmp_insta_vnf_info = copy.deepcopy(inst_vnf_info)

    inst_vnf_info.ext_cp_info = _update_ext_cp_info(change_ext_conn_req,
        inst_vnf_info=tmp_insta_vnf_info)
    inst_vnf_info.ext_virtual_link_info = _update_ext_virtual_link_info(
        change_ext_conn_req, inst_vnf_info=tmp_insta_vnf_info)

    inst_vnf_info.vnf_virtual_link_resource_info = \
        _update_vnf_virtual_link_resource_info(change_ext_conn_req,
            inst_vnf_info)

    vnf_instance.instantiated_vnf_info = inst_vnf_info


def _get_compute_nodes(vnfd_dict, instantiate_vnf_req):
    """Read the node templates and prepare VDU data in below format

    {
        'VDU1': {
            'CP': [CP1, CP2],
            'VIRTUAL_STORAGE': [virtual_storage1]
        },
    }
    """

    node_templates = vnfd_dict.get(
        'topology_template', {}).get('node_templates')

    vdu_resources = {}
    for key, value in node_templates.items():
        if value.get('type') != 'tosca.nodes.nfv.Vdu.Compute':
            continue

        desired_capacity = _convert_desired_capacity(
            instantiate_vnf_req.instantiation_level_id, vnfd_dict, key)

        cp_list = _get_cp_for_vdu(key, node_templates)

        virtual_storages = []
        requirements = value.get('requirements', [])
        for requirement in requirements:
            if requirement.get('virtual_storage'):
                virtual_storages.append(
                    requirement.get('virtual_storage'))

        vdu_resources[key] = {"CP": cp_list,
                              "VIRTUAL_STORAGE": virtual_storages,
                              "COUNT": desired_capacity}

    return vdu_resources


def _get_virtual_link_nodes(node_templates):
    virtual_link_nodes = {}

    for key, value in node_templates.items():
        if value.get('type') == 'tosca.nodes.nfv.VnfVirtualLink':
            cp_list = _get_cp_for_vl(key, node_templates)
            virtual_link_nodes[key] = cp_list

    return virtual_link_nodes


def _get_cp_for_vdu(vdu, node_templates):
    cp_list = []
    for key, value in node_templates.items():
        if value.get('type') != 'tosca.nodes.nfv.VduCp':
            continue

        requirements = value.get('requirements', [])
        for requirement in requirements:
            if requirement.get('virtual_binding') and vdu == \
                    requirement.get('virtual_binding'):
                cp_list.append(key)

    return cp_list


def _get_cp_for_vl(vl, node_templates):
    cp_list = []
    for key, value in node_templates.items():
        if value.get('type') != 'tosca.nodes.nfv.VduCp':
            continue

        requirements = value.get('requirements', [])
        for requirement in requirements:
            if requirement.get('virtual_link') and vl == \
                    requirement.get('virtual_link'):
                cp_list.append(key)

    return cp_list


def _build_vnf_virtual_link_resource_info(node_templates, instantiate_vnf_req,
                                          vnfc_resource_info, vim_id):
    virtual_link_nodes_with_cp = _get_virtual_link_nodes(node_templates)

    # Read the external networks and extcps from InstantiateVnfRequest
    for ext_virt_link in instantiate_vnf_req.ext_virtual_links:
        virtual_link_nodes_with_cp[ext_virt_link.id] = [extcp.cpd_id for extcp
                in ext_virt_link.ext_cps]

    virtual_link_resource_info_list = []

    def _get_network_resource(vl_node):
        resource_handle = objects.ResourceHandle()
        found = False
        for ext_mg_vl in instantiate_vnf_req.ext_managed_virtual_links:
            if ext_mg_vl.vnf_virtual_link_desc_id == vl_node:
                resource_handle.resource_id = ext_mg_vl.resource_id
                resource_handle.vim_connection_id = \
                    ext_mg_vl.vim_connection_id
                # TODO(tpatil): This cannot be set here.
                resource_handle.vim_level_resource_type = \
                    'OS::Neutron::Net'
                found = True
                break

        if not found:
            # check if it exists in the ext_virtual_links
            for ext_virt_link in instantiate_vnf_req.ext_virtual_links:
                if ext_virt_link.id == vl_node:
                    resource_handle.resource_id = ext_virt_link.resource_id
                    resource_handle.vim_connection_id = \
                        ext_virt_link.vim_connection_id
                    # TODO(tpatil): This cannot be set here.
                    resource_handle.vim_level_resource_type = \
                        'OS::Neutron::Net'
                    found = True
                    break

        return resource_handle

    def _get_vnf_link_port_info(cp):
        vnf_link_port_info = objects.VnfLinkPortInfo()
        vnf_link_port_info.id = uuidutils.generate_uuid()

        resource_handle = objects.ResourceHandle()
        for ext_virt_link in instantiate_vnf_req.ext_virtual_links:
            for extcp in ext_virt_link.ext_cps:
                if extcp.cpd_id == cp:
                    for cpconfig in extcp.cp_config:
                        if cpconfig.link_port_id:
                            resource_handle.resource_id = \
                                cpconfig.link_port_id
                            resource_handle.vim_connection_id = \
                                ext_virt_link.vim_connection_id
                            # TODO(tpatil): This shouldn't be set here.
                            resource_handle.vim_level_resource_type = \
                                'OS::Neutron::Port'
                            break

        vnf_link_port_info.resource_handle = resource_handle

        return vnf_link_port_info

    for node, cp_list in virtual_link_nodes_with_cp.items():
        vnf_vl_resource_info = objects.VnfVirtualLinkResourceInfo()
        vnf_vl_resource_info.id = uuidutils.generate_uuid()
        vnf_vl_resource_info.vnf_virtual_link_desc_id = node
        vnf_vl_resource_info.network_resource = _get_network_resource(node)

        vnf_link_port_info_list = []
        for cp in cp_list:
            for vnfc_resource in vnfc_resource_info:
                for vnfc_cp in vnfc_resource.vnfc_cp_info:
                    if vnfc_cp.cpd_id == cp:
                        vnf_link_port_info = _get_vnf_link_port_info(cp)
                        vnf_link_port_info.cp_instance_id = vnfc_cp.id
                        # Identifier of the "vnfLinkPorts" structure in the
                        # "vnfVirtualLinkResourceInfo" structure.
                        vnfc_cp.vnf_link_port_id = vnf_link_port_info.id
                        vnf_link_port_info_list.append(vnf_link_port_info)

        vnf_vl_resource_info.vnf_link_ports = vnf_link_port_info_list

        virtual_link_resource_info_list.append(vnf_vl_resource_info)

    return virtual_link_resource_info_list


def _update_vnf_virtual_link_resource_info(change_ext_conn_req,
                                           inst_vnf_info):
    def _update(change_ext_conn_req, vnf_vl_resource_info):
        for ext_virtual_link in change_ext_conn_req.ext_virtual_links:
            if (ext_virtual_link.id ==
                    vnf_vl_resource_info.vnf_virtual_link_desc_id):
                res_handle = objects.ResourceHandle()
                res_handle.resource_id = ext_virtual_link.resource_id
                nw_res = vnf_vl_resource_info.network_resource
                res_handle.vim_connection_id = nw_res.vim_connection_id
                res_handle.vim_level_resource_type = \
                    nw_res.vim_level_resource_type
                new_vnf_vl_resource_info = \
                    objects.VnfVirtualLinkResourceInfo(
                        id=vnf_vl_resource_info.id,
                        vnf_virtual_link_desc_id=vnf_vl_resource_info.
                        vnf_virtual_link_desc_id,
                        network_resource=res_handle,
                        vnf_link_ports=vnf_vl_resource_info.vnf_link_ports)
                return new_vnf_vl_resource_info
        return None

    vnf_virtual_link_resource_list = []
    for vnf_vl_res_info in inst_vnf_info.vnf_virtual_link_resource_info:
        updated_vnf_vl_res_info = \
            _update(change_ext_conn_req, vnf_vl_res_info)
        if updated_vnf_vl_res_info:
            vnf_virtual_link_resource_list.append(updated_vnf_vl_res_info)
        else:
            vnf_virtual_link_resource_list.append(vnf_vl_res_info)

    return vnf_virtual_link_resource_list


def _build_vnf_cp_info(instantiate_vnf_req, cp_list):
    vnfc_cp_info_list = []

    if not cp_list:
        return vnfc_cp_info_list

    def _set_vnf_exp_cp_id_protocol_data(vnfc_cp_info):
        for ext_virt_link in instantiate_vnf_req.ext_virtual_links:
            for extcp in ext_virt_link.ext_cps:
                if extcp.cpd_id == cp:
                    vnfc_cp_info.cp_protocol_info = \
                        _set_cp_protocol_info(extcp)
                    for cpconfig in extcp.cp_config:
                        vnfc_cp_info.vnf_ext_cp_id = cpconfig.link_port_id
                        break

    for cp in cp_list:
        vnfc_cp_info = objects.VnfcCpInfo()
        vnfc_cp_info.id = uuidutils.generate_uuid()
        vnfc_cp_info.cpd_id = cp
        _set_vnf_exp_cp_id_protocol_data(vnfc_cp_info)
        vnfc_cp_info_list.append(vnfc_cp_info)

    return vnfc_cp_info_list


def _build_virtual_storage_info(virtual_storages):

    for storage_node in virtual_storages:
        virtual_storage = objects.VirtualStorageResourceInfo()
        virtual_storage.id = uuidutils.generate_uuid()
        virtual_storage.virtual_storage_desc_id = storage_node

        virtual_storage.storage_resource = objects.ResourceHandle()

        yield virtual_storage


def _get_vnfc_resource_info(vnfd_dict, instantiate_vnf_req, vim_id):
    vdu_resources = _get_compute_nodes(vnfd_dict, instantiate_vnf_req)
    vnfc_resource_info_list = []
    virtual_storage_resource_info_list = []

    def _build_vnfc_resource_info(vdu, vdu_resource):
        vnfc_resource_info = objects.VnfcResourceInfo()
        vnfc_resource_info.id = uuidutils.generate_uuid()
        vnfc_resource_info.vdu_id = vdu

        vnfc_resource_info.compute_resource = objects.ResourceHandle()

        vnfc_cp_info_list = _build_vnf_cp_info(instantiate_vnf_req,
                                               vdu_resource.get("CP"))
        vnfc_resource_info.vnfc_cp_info = vnfc_cp_info_list

        virtual_storages = vdu_resource.get("VIRTUAL_STORAGE")
        vdu_storages = []
        for storage in _build_virtual_storage_info(virtual_storages):
            vdu_storages.append(storage)
            virtual_storage_resource_info_list.append(storage)

        storage_resource_ids = [info.id for info in vdu_storages]
        vnfc_resource_info.storage_resource_ids = storage_resource_ids
        return vnfc_resource_info

    for vdu, vdu_resource in vdu_resources.items():
        count = vdu_resource.get('COUNT', 1)
        for num_instance in range(count):
            vnfc_resource_info = _build_vnfc_resource_info(vdu, vdu_resource)
            vnfc_resource_info_list.append(vnfc_resource_info)

    return vnfc_resource_info_list, virtual_storage_resource_info_list


def _set_ext_cp_info(instantiate_vnf_req, inst_vnf_info=None):
    ext_cp_info_list = []
    vnfc_info = []

    if inst_vnf_info.vnfc_resource_info:
        vnfc_info = inst_vnf_info.vnfc_resource_info

    if not instantiate_vnf_req.ext_virtual_links:
        return ext_cp_info_list

    for ext_virt_link in instantiate_vnf_req.ext_virtual_links:
        if not ext_virt_link.ext_cps:
            continue

        for ext_cp in ext_virt_link.ext_cps:
            ext_cp_info = objects.VnfExtCpInfo(
                id=uuidutils.generate_uuid(),
                cpd_id=ext_cp.cpd_id,
                cp_protocol_info=_set_cp_protocol_info(ext_cp),
                ext_link_port_id=_get_ext_link_port_id(ext_virt_link,
                        ext_cp.cpd_id),
                associated_vnfc_cp_id=_get_associated_vnfc_cp_id(vnfc_info,
                        ext_cp.cpd_id))

            ext_cp_info_list.append(ext_cp_info)

    return ext_cp_info_list


def _update_ext_cp_info(change_ext_conn_req, inst_vnf_info):

    def _update(change_ext_conn_req, ext_cp_info):
        for ext_virt_link in change_ext_conn_req.ext_virtual_links:
            if not ext_virt_link.ext_cps:
                continue
            for ext_cp in ext_virt_link.ext_cps:
                if ext_cp.cpd_id == ext_cp_info.cpd_id:
                    new_ext_cp_info = objects.VnfExtCpInfo(
                        id=ext_cp_info.id,
                        cpd_id=ext_cp.cpd_id,
                        cp_protocol_info=_set_cp_protocol_info(ext_cp),
                        associated_vnfc_cp_id=ext_cp_info.
                        associated_vnfc_cp_id)
                    return new_ext_cp_info
        return None

    ext_cp_info_list = []
    for ext_cp_info in inst_vnf_info.ext_cp_info:
        updated_ext_cp_info = _update(change_ext_conn_req, ext_cp_info)
        if updated_ext_cp_info:
            ext_cp_info_list.append(updated_ext_cp_info)
        else:
            ext_cp_info_list.append(ext_cp_info)

    return ext_cp_info_list


def _get_ext_link_port_id(ext_virtual_link, cpd_id):
    if not ext_virtual_link.ext_link_ports:
        return

    for ext_link in ext_virtual_link.ext_link_ports:
        if ext_link.id == cpd_id:
            return ext_link.id


def _get_associated_vnfc_cp_id(vnfc_info, cpd_id):
    if not isinstance(vnfc_info, list):
        return

    for vnfc in vnfc_info:
        if vnfc.vnfc_cp_info:
            for cp_info in vnfc.vnfc_cp_info:
                if cp_info.cpd_id == cpd_id:
                    return vnfc.id


def _build_ip_over_ethernet_address_info(cp_protocol_data):
    """Convert IpOverEthernetAddressData to IpOverEthernetAddressInfo"""

    if not cp_protocol_data.ip_over_ethernet:
        return

    ip_over_ethernet_add_info = objects.IpOverEthernetAddressInfo()
    ip_over_ethernet_add_info.mac_address = \
        cp_protocol_data.ip_over_ethernet.mac_address

    if not cp_protocol_data.ip_over_ethernet.ip_addresses:
        return ip_over_ethernet_add_info

    ip_address_list = []
    for ip_address in cp_protocol_data.ip_over_ethernet.ip_addresses:
        ip_address_info = objects.vnf_instantiated_info.IpAddress(
            type=ip_address.type,
            addresses=ip_address.fixed_addresses,
            is_dynamic=True if ip_address.num_dynamic_addresses else False,
            subnet_id=ip_address.subnet_id)

        ip_address_list.append(ip_address_info)

    ip_over_ethernet_add_info.ip_addresses = ip_address_list

    return ip_over_ethernet_add_info


def _build_cp_protocol_info(cp_protocol_data):
    ip_over_ethernet_add_info = _build_ip_over_ethernet_address_info(
        cp_protocol_data)
    cp_protocol_info = objects.CpProtocolInfo(
        layer_protocol=cp_protocol_data.layer_protocol,
        ip_over_ethernet=ip_over_ethernet_add_info)

    return cp_protocol_info


def _set_cp_protocol_info(ext_cp):
    """Convert CpProtocolData to CpProtocolInfo"""

    cp_protocol_info_list = []
    if not ext_cp.cp_config:
        return cp_protocol_info_list

    for cp_config in ext_cp.cp_config:
        for cp_protocol_data in cp_config.cp_protocol_data:
            cp_protocol_info = _build_cp_protocol_info(cp_protocol_data)
            cp_protocol_info_list.append(cp_protocol_info)

    return cp_protocol_info_list


def _set_ext_virtual_link_info(instantiate_vnf_req, ext_cp_info):
    ext_virtual_link_list = []

    if not instantiate_vnf_req.ext_virtual_links:
        return ext_virtual_link_list

    for ext_virtual_link in instantiate_vnf_req.ext_virtual_links:
        res_handle = objects.ResourceHandle()
        res_handle.resource_id = ext_virtual_link.resource_id
        res_handle.vim_connection_id = ext_virtual_link.vim_connection_id

        ext_virtual_link_info = objects.ExtVirtualLinkInfo(
            id=ext_virtual_link.id,
            resource_handle=res_handle,
            ext_link_ports=_set_ext_link_port(ext_virtual_link,
                ext_cp_info))

        ext_virtual_link_list.append(ext_virtual_link_info)

    return ext_virtual_link_list


def _update_ext_virtual_link_info(change_ext_conn_req, inst_vnf_info):

    def _update(change_ext_conn_req, ext_virtual_link_info):
        for ext_virtual_link in change_ext_conn_req.ext_virtual_links:
            if ext_virtual_link.id == ext_virtual_link_info.id:
                res_handle = objects.ResourceHandle()
                res_handle.resource_id = ext_virtual_link.resource_id
                new_ext_virtual_link_info = objects.ExtVirtualLinkInfo(
                    id=ext_virtual_link_info.id,
                    resource_handle=res_handle,
                    ext_link_ports=ext_virtual_link_info.ext_link_ports)
                res_handle.vim_connection_id = \
                    ext_virtual_link_info.resource_handle.vim_connection_id
                new_ext_virtual_link_info = objects.ExtVirtualLinkInfo(
                    id=ext_virtual_link_info.id,
                    resource_handle=res_handle,
                    ext_link_ports=ext_virtual_link_info.ext_link_ports)
                return new_ext_virtual_link_info
        return None

    ext_virtual_link_list = []
    for ext_virtual_link_info in inst_vnf_info.ext_virtual_link_info:
        updated_ext_virtual_link_info = \
            _update(change_ext_conn_req, ext_virtual_link_info)
        if updated_ext_virtual_link_info:
            ext_virtual_link_list.append(updated_ext_virtual_link_info)
        else:
            ext_virtual_link_list.append(ext_virtual_link_info)

    return ext_virtual_link_list


def _set_ext_link_port(ext_virtual_links, ext_cp_info):
    ext_link_port_list = []

    if not ext_virtual_links.ext_link_ports:
        return ext_link_port_list

    for ext_link_port in ext_virtual_links.ext_link_ports:
        resource_handle = ext_link_port.resource_handle.obj_clone()
        cp_instance_id = None
        if ext_virtual_links.ext_cps:
            for ext_cp in ext_cp_info:
                cp_instance_id = ext_cp.id

        ext_link_port_info = objects.ExtLinkPortInfo(id=ext_link_port.id,
            resource_handle=resource_handle, cp_instance_id=cp_instance_id)

        ext_link_port_list.append(ext_link_port_info)

    return ext_link_port_list


def _build_ext_managed_virtual_link_info(instantiate_vnf_req, inst_vnf_info):

    def _network_resource(ext_managed_vl):
        resource_handle = objects.ResourceHandle(
            resource_id=ext_managed_vl.resource_id)
        # TODO(tpatil): Remove hard coding of resource type as
        # OS::Neutron::Net resource type is specific to OpenStack infra
        # driver. It could be different for other infra drivers like
        # Kubernetes.
        resource_handle.vim_level_resource_type = 'OS::Neutron::Net'
        resource_handle.vim_connection_id = \
            ext_managed_vl.vim_connection_id

        return resource_handle

    ext_managed_virtual_link_info = []
    ext_managed_virt_link_from_req = \
        instantiate_vnf_req.ext_managed_virtual_links
    for ext_managed_vl in ext_managed_virt_link_from_req:
        ext_managed_virt_info = objects.ExtManagedVirtualLinkInfo()
        ext_managed_virt_info.id = ext_managed_vl.id
        ext_managed_virt_info.vnf_virtual_link_desc_id =\
            ext_managed_vl.vnf_virtual_link_desc_id

        ext_managed_virt_info.network_resource =\
            _network_resource(ext_managed_vl)

        # Populate the vnf_link_ports from vnf_virtual_link_resource_info
        # of instantiated_vnf_info.
        for vnf_vl_res_info in inst_vnf_info.vnf_virtual_link_resource_info:
            if ext_managed_vl.vnf_virtual_link_desc_id ==\
                    vnf_vl_res_info.vnf_virtual_link_desc_id:
                vnf_link_ports = []
                for vnf_lp in vnf_vl_res_info.vnf_link_ports:
                    vnf_link_ports.append(vnf_lp.obj_clone())
                ext_managed_virt_info.vnf_link_ports = vnf_link_ports

        ext_managed_virtual_link_info.append(ext_managed_virt_info)
    return ext_managed_virtual_link_info


def _convert_desired_capacity(inst_level_id, vnfd_dict, vdu):
    aspect_delta_dict = {}
    aspect_vdu_dict = {}
    inst_level_dict = {}
    aspect_id_dict = {}
    vdu_delta_dict = {}
    aspect_max_level_dict = {}
    desired_capacity = 1

    tosca = tosca_template.ToscaTemplate(parsed_params={}, a_file=False,
                                         yaml_dict_tpl=vnfd_dict)
    tosca_policies = tosca.topology_template.policies
    default_inst_level_id = toscautils._extract_policy_info(
        tosca_policies, inst_level_dict,
        aspect_delta_dict, aspect_id_dict,
        aspect_vdu_dict, vdu_delta_dict,
        aspect_max_level_dict)

    if vdu_delta_dict.get(vdu) is None:
        return desired_capacity

    if inst_level_id:
        instantiation_level = inst_level_id
    elif default_inst_level_id:
        instantiation_level = default_inst_level_id
    else:
        return desired_capacity

    al_dict = inst_level_dict.get(instantiation_level)

    if not al_dict:
        return desired_capacity

    for aspect_id, level_num in al_dict.items():
        delta_id = aspect_id_dict.get(aspect_id)

        if delta_id is not None:
            delta_num = \
                aspect_delta_dict.get(aspect_id).get(delta_id)

        vdus = aspect_vdu_dict.get(aspect_id)
        initial_delta = None
        if vdu in vdus:
            initial_delta = vdu_delta_dict.get(vdu)

        if initial_delta is not None:
            desired_capacity = initial_delta + delta_num * level_num

    return desired_capacity


def _get_vnf_package_path(context, vnfd_id):
    return os.path.join(CONF.vnf_package.vnf_package_csar_path,
                        _get_vnf_package_id(context, vnfd_id))


def get_base_nest_hot_dict(context, flavour_id, vnfd_id):
    base_hot_path = os.path.join(_get_vnf_package_path(context, vnfd_id),
                                 'BaseHOT', flavour_id)
    nested_hot_path = os.path.join(base_hot_path, 'nested')
    ext = (".yaml", ".yml")

    base_hot_dict = None
    nested_hot_dict = {}
    if os.path.exists(base_hot_path):
        for file in os.listdir(base_hot_path):
            if file.endswith(ext):
                with open(os.path.join(base_hot_path, file)) as file_obj:
                    base_hot_dict = yaml.safe_load(file_obj)
    if os.path.exists(nested_hot_path):
        for file in os.listdir(nested_hot_path):
            if file.endswith(ext):
                with open(os.path.join(nested_hot_path, file)) as file_obj:
                    nested_hot = yaml.safe_load(file_obj)
                    nested_hot_dict[file] = nested_hot
    LOG.debug("Loaded base hot: %s", base_hot_dict)
    LOG.debug("Loaded nested_hot_dict: %s", nested_hot_dict)
    return base_hot_dict, nested_hot_dict


def get_extract_policy_infos(tosca):
    aspect_delta_dict = {}
    aspect_vdu_dict = {}
    inst_level_dict = {}
    aspect_id_dict = {}
    vdu_delta_dict = {}
    aspect_max_level_dict = {}

    tosca_policies = tosca.topology_template.policies
    default_inst_level_id = toscautils._extract_policy_info(
        tosca_policies, inst_level_dict,
        aspect_delta_dict, aspect_id_dict,
        aspect_vdu_dict, vdu_delta_dict,
        aspect_max_level_dict)

    extract_policy_infos = dict()
    extract_policy_infos['inst_level_dict'] = inst_level_dict
    extract_policy_infos['aspect_delta_dict'] = aspect_delta_dict
    extract_policy_infos['aspect_id_dict'] = aspect_id_dict
    extract_policy_infos['aspect_vdu_dict'] = aspect_vdu_dict
    extract_policy_infos['vdu_delta_dict'] = vdu_delta_dict
    extract_policy_infos['aspect_max_level_dict'] = aspect_max_level_dict
    extract_policy_infos['default_inst_level_id'] = default_inst_level_id

    return extract_policy_infos


def get_scale_delta_num(extract_policy_infos, aspect_id):
    delta_num = 1

    if extract_policy_infos['aspect_id_dict'] is None:
        return delta_num
    delta_id = extract_policy_infos['aspect_id_dict'].get(aspect_id)
    if delta_id is None:
        return delta_num
    delta_num = \
        extract_policy_infos['aspect_delta_dict'].get(aspect_id).get(delta_id)

    return delta_num


def get_default_scale_status(context, vnf_instance, vnfd_dict):
    default_scale_status = None

    vnfd_dict = _get_vnfd_dict(context,
        vnf_instance.vnfd_id,
        vnf_instance.instantiated_vnf_info.flavour_id)
    tosca = tosca_template.ToscaTemplate(parsed_params={}, a_file=False,
                                         yaml_dict_tpl=vnfd_dict)
    extract_policy_infos = get_extract_policy_infos(tosca)

    if extract_policy_infos['inst_level_dict'] is None:
        return default_scale_status
    default_inst_level_id = extract_policy_infos['default_inst_level_id']
    default_al_dict = \
        extract_policy_infos['inst_level_dict'].get(default_inst_level_id)
    if default_al_dict is None:
        return default_scale_status
    default_scale_status = []
    for aspect_id, level_num in default_al_dict.items():
        default_scale_status.append(
            objects.ScaleInfo(
                aspect_id=aspect_id,
                scale_level=level_num))

    return default_scale_status


def get_target_vdu_def_dict(extract_policy_infos, aspect_id, tosca):
    vdu_def_dict = {}

    tosca_node_tpls = tosca.topology_template.nodetemplates

    if extract_policy_infos['aspect_vdu_dict'] is None:
        return vdu_def_dict
    vdus = extract_policy_infos['aspect_vdu_dict'].get(aspect_id)
    if vdus is None:
        return vdu_def_dict
    for nt in tosca_node_tpls:
        for node_name, node_value in nt.templates.items():
            if node_name in vdus:
                vdu_def_dict[node_name] = node_value

    return vdu_def_dict


def _get_changed_ext_connectivity(
        old_vnf_instance=None, new_vnf_instance=None):

    changed_ext_connectivities = []
    if not old_vnf_instance or not new_vnf_instance:
        return changed_ext_connectivities

    old_vnf_vl_res_info = \
        old_vnf_instance.instantiated_vnf_info.\
        vnf_virtual_link_resource_info
    new_vnf_vl_res_info = \
        new_vnf_instance.instantiated_vnf_info.\
        vnf_virtual_link_resource_info

    def _compare_vnf_link_ports(old_vnf_link_ports,
                                new_vnf_link_ports):
        differed_vnf_link_ports = []
        for old_vnf_link_port in old_vnf_link_ports:
            for new_vnf_link_port in new_vnf_link_ports:
                if old_vnf_link_port.id == new_vnf_link_port.id:
                    if (old_vnf_link_port.resource_handle.resource_id !=
                            new_vnf_link_port.resource_handle.resource_id):
                        differed_vnf_link_ports.append(new_vnf_link_port)
        return differed_vnf_link_ports

    for old_vl_res in old_vnf_vl_res_info:
        for new_vl_res in new_vnf_vl_res_info:
            if old_vl_res.id == new_vl_res.id:
                changed_ext_connectivity = objects.ExtVirtualLinkInfo(
                    id=new_vl_res.id,
                    resource_handle=new_vl_res.network_resource,
                    ext_link_ports=[])
                differed_vnf_link_ports = _compare_vnf_link_ports(
                    old_vl_res.vnf_link_ports,
                    new_vl_res.vnf_link_ports)
                for link_port in differed_vnf_link_ports:
                    changed_ext_link_port = objects.ExtLinkPortInfo(
                        id=link_port.id,
                        resource_handle=link_port.resource_handle,
                        cp_instance_id=link_port.cp_instance_id)
                    changed_ext_connectivity.ext_link_ports.\
                        append(changed_ext_link_port)
                if changed_ext_connectivity.ext_link_ports:
                    changed_ext_connectivities.append(
                        changed_ext_connectivity)

    LOG.debug('changed_ext_connectivities: {}'.format(
        changed_ext_connectivities))
    return changed_ext_connectivities



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import tacker.vnfm.lcm_user_data.utils as UserDataUtil

from tacker.vnfm.lcm_user_data.abstract_user_data import AbstractUserData


class SampleUserData(AbstractUserData):

    @staticmethod
    def instantiate(base_hot_dict=None,
                    vnfd_dict=None,
                    inst_req_info=None,
                    grant_info=None):

        # Create HOT input parameter using util functions.
        initial_param_dict = UserDataUtil.create_initial_param_dict(
            base_hot_dict)

        vdu_flavor_dict = UserDataUtil.create_vdu_flavor_dict(vnfd_dict)
        vdu_image_dict = UserDataUtil.create_vdu_image_dict(grant_info)
        cpd_vl_dict = UserDataUtil.create_cpd_vl_dict(
            base_hot_dict, inst_req_info)

        final_param_dict = UserDataUtil.create_final_param_dict(
            initial_param_dict, vdu_flavor_dict, vdu_image_dict, cpd_vl_dict)

        return final_param_dict



# Copyright 2012 New Dream Network, LLC (DreamHost)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from logging import config as logging_config

from alembic import context
from sqlalchemy import create_engine, pool

from tacker.db import model_base


# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config
tacker_config = config.tacker_config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
logging_config.fileConfig(config.config_file_name)

# set the target for 'autogenerate' support
target_metadata = model_base.BASE.metadata


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with either a URL
    or an Engine.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    kwargs = dict()
    if tacker_config.database.connection:
        kwargs['url'] = tacker_config.database.connection
    else:
        kwargs['dialect_name'] = tacker_config.database.engine
    context.configure(**kwargs)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    engine = create_engine(
        tacker_config.database.connection,
        poolclass=pool.NullPool)

    connection = engine.connect()
    context.configure(
        connection=connection,
        target_metadata=target_metadata
    )

    try:
        with context.begin_transaction():
            context.run_migrations()
    finally:
        connection.close()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from tacker.vnfm.lcm_user_data.abstract_user_data import AbstractUserData
import tacker.vnfm.lcm_user_data.utils as UserDataUtil


class SampleUserData(AbstractUserData):
    @staticmethod
    def instantiate(base_hot_dict=None,
                    vnfd_dict=None,
                    inst_req_info=None,
                    grant_info=None):
        api_param = UserDataUtil.get_diff_base_hot_param_from_api(
            base_hot_dict, inst_req_info)
        initial_param_dict = \
            UserDataUtil.create_initial_param_server_port_dict(
                base_hot_dict)
        vdu_flavor_dict = \
            UserDataUtil.create_vdu_flavor_capability_name_dict(vnfd_dict)
        vdu_image_dict = UserDataUtil.create_sw_image_dict(vnfd_dict)
        cpd_vl_dict = UserDataUtil.create_network_dict(
            inst_req_info, initial_param_dict)
        final_param_dict = UserDataUtil.create_final_param_dict(
            initial_param_dict, vdu_flavor_dict, vdu_image_dict, cpd_vl_dict)
        return {**final_param_dict, **api_param}



# Copyright (C) 2020 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from kubernetes import client
import os
from unittest import mock

from tacker.common import exceptions
from tacker.tests.unit import base
from tacker.tests.unit import fake_request
from tacker.tests.unit.vnfm.infra_drivers.kubernetes import fakes
from tacker.vnfm.infra_drivers.kubernetes.k8s import tosca_kube_object
from tacker.vnfm.infra_drivers.kubernetes.k8s import translate_outputs


class TestTransformer(base.TestCase):
    def setUp(self):
        super(TestTransformer, self).setUp()
        self.yaml_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "kubernetes_api_resource/")
        self.k8s_client_dict = fakes.fake_k8s_client_dict()
        self.transfromer = translate_outputs.Transformer(
            client.CoreV1Api,
            client.AppsV1Api,
            client.AutoscalingApi,
            self.k8s_client_dict)

    def test_deploy_k8s_create_false(self):
        kubernetes_objects = []
        k8s_obj = fakes.fake_k8s_dict()
        kubernetes_objects.append(k8s_obj)
        self.assertRaises(exceptions.CreateApiFalse,
                          self.transfromer.deploy_k8s,
                          kubernetes_objects)

    @mock.patch.object(translate_outputs.Transformer,
                       "_select_k8s_client_and_api")
    def test_deploy_k8s(self, mock_k8s_client_and_api):
        req = fake_request.HTTPRequest.blank(
            'apis/apps/v1/namespaces/curryns/deployments')
        mock_k8s_client_and_api.return_value = req
        kubernetes_objects = []
        k8s_obj = fakes.fake_k8s_dict()
        kubernetes_objects.append(k8s_obj)
        new_k8s_objs = self.transfromer.deploy_k8s(kubernetes_objects)
        self.assertIsInstance(new_k8s_objs, list)
        self.assertIsNotNone(new_k8s_objs)
        self.assertEqual(new_k8s_objs[0]['status'], 'Creating')

    def test_deployment(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['deployment.yaml'], self.yaml_path)
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind, 'Deployment')
        self.assertEqual(k8s_objs[0].get('object').api_version, 'apps/v1')

    def test_api_service(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['api-service.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind, 'APIService')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'apiregistration.k8s.io/v1')

    def test_cluster_role(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['cluster-role.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind, 'ClusterRole')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'rbac.authorization.k8s.io/v1')

    def test_cluster_role_binding(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['cluster-role-binding.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'ClusterRoleBinding')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'rbac.authorization.k8s.io/v1')

    def test_config_map(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['config-map.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'curryns')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'ConfigMap')
        self.assertEqual(k8s_objs[0].get('object').api_version, 'v1')

    def test_daemon_set(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['daemon-set.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'DaemonSet')
        self.assertEqual(k8s_objs[0].get('object').api_version, 'apps/v1')

    def test_horizontal_pod_autoscaler(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['horizontal-pod-autoscaler.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'default')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'HorizontalPodAutoscaler')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'autoscaling/v1')

    def test_job(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['job.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind, 'Job')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'batch/v1')

    def test_lease(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['lease.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'default')
        self.assertEqual(k8s_objs[0].get('object').kind, 'Lease')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'coordination.k8s.io/v1')

    def test_local_subject_access_review(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['local-subject-access-review.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'curry-ns')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'LocalSubjectAccessReview')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'authorization.k8s.io/v1')

    def test_namespace(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['namespace.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind, 'Namespace')
        self.assertEqual(k8s_objs[0].get('object').api_version, 'v1')

    def test_network_policy(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['network-policy.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind, 'NetworkPolicy')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'networking.k8s.io/v1')

    def test_node(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['node.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind, 'Node')
        self.assertEqual(k8s_objs[0].get('object').api_version, 'v1')

    def test_persistent_volume(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['persistent-volume.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind, 'PersistentVolume')
        self.assertEqual(k8s_objs[0].get('object').api_version, 'v1')

    def test_persistent_volume_claim(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['persistent-volume-claim.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'PersistentVolumeClaim')
        self.assertEqual(k8s_objs[0].get('object').api_version, 'v1')

    def test_pod(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['pod.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'Pod')
        self.assertEqual(k8s_objs[0].get('object').api_version, 'v1')

    def test_priority_class(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['priority-class.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'PriorityClass')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'scheduling.k8s.io/v1')

    def test_replica_set(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['replica-set.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'ReplicaSet')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'apps/v1')

    def test_resource_quota(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['resource-quota.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'curryns')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'ResourceQuota')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'v1')

    def test_role(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['role.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'curry-ns')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'Role')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'rbac.authorization.k8s.io/v1')

    def test_role_binding(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['role-bindings.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'curry-ns')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'RoleBinding')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'rbac.authorization.k8s.io/v1')

    def test_secret(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['secret.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'default')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'Secret')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'v1')

    def test_self_subject_access_review(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['self-subject-access-review.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'SelfSubjectAccessReview')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'authorization.k8s.io/v1')

    def test_self_subject_rules_review(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['self-subject-rule-review.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'SelfSubjectRulesReview')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'authorization.k8s.io/v1')

    def test_service(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['service.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'default')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'Service')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'v1')

    def test_service_account(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['service-account.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'default')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'ServiceAccount')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'v1')

    def test_stateful_set(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['stateful-set.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'StatefulSet')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'apps/v1')

    def test_storage_class(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['storage-class.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'StorageClass')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'storage.k8s.io/v1')

    def test_subject_access_review(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['subject-access-review.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'SubjectAccessReview')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'authorization.k8s.io/v1')

    def test_token_review(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['token-review.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), '')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'TokenReview')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'authentication.k8s.io/v1')

    def test_limit_range(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['limit-range.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'curryns')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'LimitRange')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'v1')

    def test_pod_template(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['pod-template.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'curryns')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'PodTemplate')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'v1')

    def test_volume_attachment(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['volume-attachment.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'curryns')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'VolumeAttachment')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'storage.k8s.io/v1')

    def test_bindings(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['bindings.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'curryns')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'Binding')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'v1')

    def test_controller_revision(self):
        k8s_objs = self.transfromer.get_k8s_objs_from_yaml(
            ['controller-revision.yaml'], self.yaml_path
        )
        self.assertIsNotNone(k8s_objs[0].get('object'))
        self.assertEqual(k8s_objs[0].get('namespace'), 'curryns')
        self.assertEqual(k8s_objs[0].get('object').kind,
                         'ControllerRevision')
        self.assertEqual(k8s_objs[0].get('object').api_version,
                         'apps/v1')

    def test_transform(self):
        container_obj = tosca_kube_object.Container(
            config='config:abc\nconfig2:bcd',
            num_cpus=2,
            mem_size=10,
            name='container'
        )
        tosca_kube_objects = [tosca_kube_object.ToscaKubeObject(
            namespace='namespace',
            name='name',
            containers=[container_obj],
            mapping_ports=["123"],
            labels={}
        )]
        kubernetes_objects = self.transfromer.transform(tosca_kube_objects)
        self.assertEqual(kubernetes_objects['namespace'], 'namespace')
        self.assertEqual(
            kubernetes_objects['objects'][0].data, {
                'config': 'abc', 'config2': 'bcd'})

    @mock.patch.object(client.CoreV1Api, 'create_namespaced_config_map')
    @mock.patch.object(client.AppsV1Api, 'create_namespaced_deployment')
    @mock.patch.object(client.CoreV1Api, 'create_namespaced_service')
    def test_deploy(
            self,
            mock_create_namespaced_config_map,
            mock_create_namespaced_deployment,
            mock_create_namespaced_service):
        mock_create_namespaced_config_map.return_value = ""
        mock_create_namespaced_deployment.return_value = ""
        mock_create_namespaced_service.return_value = ""
        container_obj = tosca_kube_object.Container(
            config='config:abc\nconfig2:bcd',
            num_cpus=2,
            mem_size=10,
            name='container'
        )
        tosca_kube_objects = [tosca_kube_object.ToscaKubeObject(
            namespace='namespace',
            name='name',
            containers=[container_obj],
            mapping_ports=["123"],
            labels={}
        )]
        kubernetes_objects = self.transfromer.transform(tosca_kube_objects)
        result = self.transfromer.deploy(kubernetes_objects)
        self.assertEqual(result, 'namespace,name')



# Copyright (C) 2020 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import time

from oslo_serialization import jsonutils
from oslo_utils import uuidutils
from sqlalchemy import desc
from sqlalchemy.orm import joinedload

from tacker.common import exceptions
from tacker import context
from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.db_sqlalchemy import models
from tacker.objects import fields
from tacker.objects import vnf_lcm_op_occs
from tacker.tests.functional import base
from tacker.tests import utils

VNF_PACKAGE_UPLOAD_TIMEOUT = 300
VNF_INSTANTIATE_TIMEOUT = 600
VNF_TERMINATE_TIMEOUT = 600
VNF_HEAL_TIMEOUT = 600
RETRY_WAIT_TIME = 5


def _create_and_upload_vnf_package(tacker_client, csar_package_name,
                                   user_defined_data):
    # create vnf package
    body = jsonutils.dumps({"userDefinedData": user_defined_data})
    resp, vnf_package = tacker_client.do_request(
        '/vnfpkgm/v1/vnf_packages', "POST", body=body)

    # upload vnf package
    csar_package_path = "../../../etc/samples/etsi/nfv/%s" % csar_package_name
    file_path = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                             csar_package_path))

    # Generating unique vnfd id. This is required when multiple workers
    # are running concurrently. The call below creates a new temporary
    # CSAR with unique vnfd id.
    file_path, uniqueid = utils.create_csar_with_unique_vnfd_id(file_path)

    with open(file_path, 'rb') as file_object:
        resp, resp_body = tacker_client.do_request(
            '/vnfpkgm/v1/vnf_packages/{id}/package_content'.format(
                id=vnf_package['id']),
            "PUT", body=file_object, content_type='application/zip')

    # wait for onboard
    start_time = int(time.time())
    show_url = os.path.join('/vnfpkgm/v1/vnf_packages', vnf_package['id'])
    vnfd_id = None
    while True:
        resp, body = tacker_client.do_request(show_url, "GET")
        if body['onboardingState'] == "ONBOARDED":
            vnfd_id = body['vnfdId']
            break

        if ((int(time.time()) - start_time) > VNF_PACKAGE_UPLOAD_TIMEOUT):
            raise Exception("Failed to onboard vnf package, process could not"
                " be completed within %d seconds", VNF_PACKAGE_UPLOAD_TIMEOUT)

        time.sleep(RETRY_WAIT_TIME)

    # remove temporarily created CSAR file
    os.remove(file_path)
    return vnf_package['id'], vnfd_id


def _delete_wait_vnf_instance(tacker_client, id):
    url = os.path.join("/vnflcm/v1/vnf_instances", id)
    start_time = int(time.time())
    while True:
        resp, body = tacker_client.do_request(url, "DELETE")
        if 204 == resp.status_code:
            break

        if ((int(time.time()) - start_time) > VNF_TERMINATE_TIMEOUT):
            raise Exception("Failed to delete vnf instance, process could not"
                " be completed within %d seconds", VNF_TERMINATE_TIMEOUT)

        time.sleep(RETRY_WAIT_TIME)


def _show_vnf_instance(tacker_client, id):
    show_url = os.path.join("/vnflcm/v1/vnf_instances", id)
    resp, vnf_instance = tacker_client.do_request(show_url, "GET")

    return vnf_instance


def _vnf_instance_wait(
        tacker_client, id,
        instantiation_state=fields.VnfInstanceState.INSTANTIATED,
        timeout=VNF_INSTANTIATE_TIMEOUT):
    show_url = os.path.join("/vnflcm/v1/vnf_instances", id)
    start_time = int(time.time())
    while True:
        resp, body = tacker_client.do_request(show_url, "GET")
        if body['instantiationState'] == instantiation_state:
            break

        if ((int(time.time()) - start_time) > timeout):
            raise Exception("Failed to wait vnf instance, process could not"
                " be completed within %d seconds", timeout)

        time.sleep(RETRY_WAIT_TIME)


class VnfLcmKubernetesHealTest(base.BaseTackerTest):

    @classmethod
    def setUpClass(cls):
        cls.tacker_client = base.BaseTackerTest.tacker_http_client()
        cls.vnf_package_resource, cls.vnfd_id_resource = \
            _create_and_upload_vnf_package(
                cls.tacker_client, "test_cnf_heal",
                {"key": "sample_heal_functional"})
        cls.vnf_instance_ids = []
        super(VnfLcmKubernetesHealTest, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        # Update vnf package operational state to DISABLED
        update_req_body = jsonutils.dumps({
            "operationalState": "DISABLED"})
        base_path = "/vnfpkgm/v1/vnf_packages"
        for package_id in [cls.vnf_package_resource]:
            resp, resp_body = cls.tacker_client.do_request(
                '{base_path}/{id}'.format(id=package_id,
                                          base_path=base_path),
                "PATCH", content_type='application/json',
                body=update_req_body)

            # Delete vnf package
            url = '/vnfpkgm/v1/vnf_packages/%s' % package_id
            cls.tacker_client.do_request(url, "DELETE")

        super(VnfLcmKubernetesHealTest, cls).tearDownClass()

    def setUp(self):
        super(VnfLcmKubernetesHealTest, self).setUp()
        self.base_vnf_instances_url = "/vnflcm/v1/vnf_instances"
        self.base_vnf_lcm_op_occs_url = "/vnflcm/v1/vnf_lcm_op_occs"
        self.context = context.get_admin_context()
        vim_list = self.client.list_vims()
        if not vim_list:
            self.skipTest("Vims are not configured")

        vim_id = 'vim-kubernetes'
        vim = self.get_vim(vim_list, vim_id)
        if not vim:
            self.skipTest("Kubernetes VIM '%s' is missing" % vim_id)
        self.vim_id = vim['id']

    def _instantiate_vnf_instance_request(
            self, flavour_id, vim_id=None, additional_param=None):
        request_body = {"flavourId": flavour_id}

        if vim_id:
            request_body["vimConnectionInfo"] = [
                {"id": uuidutils.generate_uuid(),
                 "vimId": vim_id,
                 "vimType": "kubernetes"}]

        if additional_param:
            request_body["additionalParams"] = additional_param

        return request_body

    def _create_vnf_instance(self, vnfd_id, vnf_instance_name=None,
                             vnf_instance_description=None):
        request_body = {'vnfdId': vnfd_id}
        if vnf_instance_name:
            request_body['vnfInstanceName'] = vnf_instance_name

        if vnf_instance_description:
            request_body['vnfInstanceDescription'] = vnf_instance_description

        resp, response_body = self.http_client.do_request(
            self.base_vnf_instances_url, "POST",
            body=jsonutils.dumps(request_body))
        return resp, response_body

    def _instantiate_vnf_instance(self, id, request_body):
        url = os.path.join(self.base_vnf_instances_url, id, "instantiate")
        resp, body = self.http_client.do_request(
            url, "POST", body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)
        _vnf_instance_wait(self.tacker_client, id)

    def _create_and_instantiate_vnf_instance(self, flavour_id,
                                             additional_params):
        # create vnf instance
        vnf_instance_name = "test_vnf_instance_for_cnf_heal-%s" % \
                            uuidutils.generate_uuid()
        vnf_instance_description = "vnf instance for cnf heal testing"
        resp, vnf_instance = self._create_vnf_instance(
            self.vnfd_id_resource, vnf_instance_name=vnf_instance_name,
            vnf_instance_description=vnf_instance_description)

        # instantiate vnf instance
        additional_param = additional_params
        request_body = self._instantiate_vnf_instance_request(
            flavour_id, vim_id=self.vim_id, additional_param=additional_param)

        self._instantiate_vnf_instance(vnf_instance['id'], request_body)
        vnf_instance = _show_vnf_instance(
            self.tacker_client, vnf_instance['id'])
        self.vnf_instance_ids.append(vnf_instance['id'])

        return vnf_instance

    def _terminate_vnf_instance(self, id):
        # Terminate vnf forcefully
        request_body = {
            "terminationType": fields.VnfInstanceTerminationType.FORCEFUL,
        }
        url = os.path.join(self.base_vnf_instances_url, id, "terminate")
        resp, body = self.http_client.do_request(
            url, "POST", body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)
        _vnf_instance_wait(
            self.tacker_client, id,
            instantiation_state=fields.VnfInstanceState.NOT_INSTANTIATED,
            timeout=VNF_TERMINATE_TIMEOUT)

    def _delete_vnf_instance(self, id):
        _delete_wait_vnf_instance(self.tacker_client, id)

        # verify vnf instance is deleted
        url = os.path.join(self.base_vnf_instances_url, id)
        resp, body = self.http_client.do_request(url, "GET")
        self.assertEqual(404, resp.status_code)

    def _heal_vnf_instance(self, id, vnfc_instance_id):
        url = os.path.join(self.base_vnf_instances_url, id, "heal")
        # generate body
        request_body = {
            "vnfcInstanceId": vnfc_instance_id}
        resp, body = self.http_client.do_request(
            url, "POST", body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)

    @db_api.context_manager.reader
    def _vnf_notify_get_by_id(self, context, vnf_instance_id,
                              columns_to_join=None):
        query = api.model_query(
            context, models.VnfLcmOpOccs,
            read_deleted="no", project_only=True).filter_by(
            vnf_instance_id=vnf_instance_id).order_by(
            desc("created_at"))

        if columns_to_join:
            for column in columns_to_join:
                query = query.options(joinedload(column))

        db_vnflcm_op_occ = query.first()

        if not db_vnflcm_op_occ:
            raise exceptions.VnfInstanceNotFound(id=vnf_instance_id)

        vnflcm_op_occ = vnf_lcm_op_occs.VnfLcmOpOcc.obj_from_db_obj(
            context, db_vnflcm_op_occ)
        return vnflcm_op_occ

    def _wait_vnflcm_op_occs(
            self, context, vnf_instance_id,
            operation_state='COMPLETED'):
        start_time = int(time.time())
        while True:
            vnflcm_op_occ = self._vnf_notify_get_by_id(
                context, vnf_instance_id)

            if vnflcm_op_occ.operation_state == operation_state:
                break

            if ((int(time.time()) - start_time) > VNF_HEAL_TIMEOUT):
                raise Exception("Failed to wait heal instance")

            time.sleep(RETRY_WAIT_TIME)

    def _get_vnfc_resource_info(self, vnf_instance):
        inst_vnf_info = vnf_instance['instantiatedVnfInfo']
        vnfc_resource_info = inst_vnf_info['vnfcResourceInfo']
        return vnfc_resource_info

    def test_heal_cnf_with_sol002(self):
        """Test heal as per SOL002 for CNF

        This test will instantiate cnf. Heal API will be invoked as per SOL002
        i.e. with vnfcInstanceId, so that the specified vnfc instance is healed
        which includes Kubernetes resources (Pod and Deployment).
        """
        # use def-files of singleton Pod and Deployment (replicas=2)
        inst_additional_param = {
            "lcm-kubernetes-def-files": [
                "Files/kubernetes/deployment_heal_complex.yaml",
                "Files/kubernetes/pod_heal.yaml"]}
        vnf_instance = self._create_and_instantiate_vnf_instance(
            "complex", inst_additional_param)
        before_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)

        # get vnfc_instance_id of heal target
        deployment_target_vnfc = None
        for vnfc_rsc in before_vnfc_rscs:
            compute_resource = vnfc_rsc['computeResource']
            rsc_kind = compute_resource['vimLevelResourceType']
            if rsc_kind == 'Pod':
                # target 1: Singleton Pod
                pod_target_vnfc = vnfc_rsc
            elif not deployment_target_vnfc:
                # target 2: Deployment's Pod
                deployment_target_vnfc = vnfc_rsc
            else:
                # not target: Deployment's remianing one
                deployment_not_target_vnfc = vnfc_rsc

        # test heal SOL-002 (partial heal)
        vnfc_instance_id = \
            [pod_target_vnfc['id'], deployment_target_vnfc['id']]
        self._heal_vnf_instance(vnf_instance['id'], vnfc_instance_id)
        # wait vnflcm_op_occs.operation_state become COMPLETE
        self._wait_vnflcm_op_occs(self.context, vnf_instance['id'])
        # check vnfcResourceInfo after heal operation
        vnf_instance = _show_vnf_instance(
            self.tacker_client, vnf_instance['id'])
        after_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)
        self.assertEqual(len(before_vnfc_rscs), len(after_vnfc_rscs))
        for vnfc_rsc in after_vnfc_rscs:
            after_pod_name = vnfc_rsc['computeResource']['resourceId']
            if vnfc_rsc['id'] == pod_target_vnfc['id']:
                # check stored pod name is not changed (Pod)
                after_resource = pod_target_vnfc
                compute_resource = after_resource['computeResource']
                before_pod_name = compute_resource['resourceId']
                self.assertEqual(after_pod_name, before_pod_name)
            elif vnfc_rsc['id'] == deployment_target_vnfc['id']:
                # check stored pod name is changed (Deployment)
                after_resource = deployment_target_vnfc
                compute_resource = after_resource['computeResource']
                before_pod_name = compute_resource['resourceId']
                self.assertNotEqual(after_pod_name, before_pod_name)
            else:
                # check stored pod name is not changed (not target)
                after_resource = deployment_not_target_vnfc
                compute_resource = after_resource['computeResource']
                before_pod_name = compute_resource['resourceId']
                self.assertEqual(after_pod_name, before_pod_name)
        self._terminate_vnf_instance(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])

    def test_heal_cnf_with_sol003(self):
        """Test heal as per SOL003 for CNF

        This test will instantiate cnf. Heal API will be invoked as per SOL003
        i.e. without passing vnfcInstanceId, so that the entire vnf is healed
        which includes Kubernetes resource (Deployment).
        """
        # use def-files of Deployment (replicas=2)
        inst_additional_param = {
            "lcm-kubernetes-def-files": [
                "Files/kubernetes/deployment_heal_simple.yaml"]}
        vnf_instance = self._create_and_instantiate_vnf_instance(
            "simple", inst_additional_param)
        before_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)

        # test heal SOL-003 (entire heal)
        vnfc_instance_id = []
        self._heal_vnf_instance(vnf_instance['id'], vnfc_instance_id)
        # wait vnflcm_op_occs.operation_state become COMPLETE
        self._wait_vnflcm_op_occs(self.context, vnf_instance['id'])
        # check vnfcResourceInfo after heal operation
        vnf_instance = _show_vnf_instance(
            self.tacker_client, vnf_instance['id'])
        after_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)
        self.assertEqual(len(before_vnfc_rscs), len(after_vnfc_rscs))
        # check id and pod name (as computeResource.resourceId) is changed
        for before_vnfc_rsc in before_vnfc_rscs:
            for after_vnfc_rsc in after_vnfc_rscs:
                self.assertNotEqual(
                    before_vnfc_rsc['id'], after_vnfc_rsc['id'])
                self.assertNotEqual(
                    before_vnfc_rsc['computeResource']['resourceId'],
                    after_vnfc_rsc['computeResource']['resourceId'])
        # terminate vnf instance
        self._terminate_vnf_instance(vnf_instance['id'])
        self._delete_vnf_instance(vnf_instance['id'])



# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import fixtures
from oslo_log import log as logging

from tacker.agent.linux import utils
from tacker.tests import base


LOG = logging.getLogger(__name__)


class RootwrapTestExec(base.BaseTestCase):
    """Simple unit test to test the basic rootwrap mechanism

    Essentially hello-world.  Just run a command as root and check that
    it actually *did* run as root, and generated the right output.

    NB that this is named _test_rootwrap so as not to get run by default
    from scripts like tox.  That's because it actually executes a sudo'ed
    command, and that won't work in the automated test environment, at
    least as it stands today.  To run this, rename it to
    test_rootwrap.py, or run it by hand.
    """

    def setUp(self):
        super(RootwrapTestExec, self).setUp()
        self.cwd = os.getcwd() + "/../../.."
        # stuff a stupid bash script into /tmp, so that the next
        # method can execute it.
        self.test_file = self.useFixture(
            fixtures.TempDir()).join("rootwrap-test.sh")
        with open(self.test_file, 'w') as f:
            f.write('#!/bin/bash\n')
            f.write('ID=`id | sed \'s/uid=//\' | sed \'s/(.*//\' `\n')
            f.write("echo $ID $1\
\" Now is the time for all good men to come \
to the aid of their party.\"\n")
        # we need a temporary conf file, pointing into pwd for the filter
        # specs. there's probably a better way to do this, but I couldn't
        # figure it out.  08/15/12 -- jrd
        self.conf_file = self.useFixture(
            fixtures.TempDir()).join("rootwrap.conf")
        with open(self.conf_file, 'w') as f:
            f.write("# temporary conf file for rootwrap-test, " +
                    "generated by test_rootwrap.py\n")
            f.write("[DEFAULT]\n")
            f.write("filters_path=" + self.cwd +
                    "/tacker/tests/etc/rootwrap.d/")
        # now set the root helper to sudo our rootwrap script,
        #  with the new conf
        self.root_helper = "sudo " + self.cwd + "/bin/tacker-rootwrap "
        self.root_helper += self.conf_file

    def runTest(self):
        try:
            result = utils.execute(["bash", self.test_file, 'arg'],
                                   self.root_helper)
            self.assertEqual("0 arg Now is the time for all good men to \
come to the aid of their party.", result)
        except Exception:
            LOG.exception("Losing in rootwrap test")

    def tearDown(self):
        os.remove(self.test_file)
        os.remove(self.conf_file)
        super(RootwrapTestExec, self).tearDown()



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


CONF = cfg.CONF

interval_opts = [
    cfg.IntOpt('vnf_package_delete_interval',
               default=1800,
               help=_('Seconds between running periodic tasks '
                      'to cleanup residues of deleted vnf packages')),
]


def register_opts(conf):
    conf.register_opts(interval_opts)


def list_opts():
    return {'DEFAULT': interval_opts}



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
from oslo_serialization import jsonutils
from tacker.tests import uuidsentinel


class Grant:
    GRANT_REQ_PATH = '/grant/v1/grants'

    ZONES = [
        {
            "id": uuidsentinel.zone_id,
            "zoneId": "nova",
            "vimConnectionId": uuidsentinel.vim_connection_id
        }
    ]

    ADDITIONAL_PARAMS = {
        "key": "value"
    }

    @staticmethod
    def _make_vim_connection_info(tenant_id):
        access_info = {
            "username": "nfv_user",
            "region": "RegionOne",
            "password": "devstack",
            "tenant": tenant_id
        }

        return [{
            "id": uuidsentinel.vim_connection_id,
            "vimType": "ETSINFV.OPENSTACK_KEYSTONE.v_2",
            "interfaceInfo": {
                "endpoint": "http://127.0.0.1/identity"
            },
            "accessInfo": access_info
        }]

    @staticmethod
    def _make_add_resources(req_add_resources):
        add_resources = []
        for req_add_resource in req_add_resources:
            res_add_resource = {
                "resourceDefinitionId": req_add_resource['id'],
                "vimConnectionId": uuidsentinel.vim_connection_id
            }

            if req_add_resource['type'] == 'COMPUTE':
                res_add_resource["zoneId"] = uuidsentinel.zone_id

            add_resources.append(res_add_resource)

        return add_resources

    @staticmethod
    def _make_remove_resources(req_remove_resources):
        res_remove_resources = []
        for req_remove_resource in req_remove_resources:
            res_remove_resource = {
                "resourceDefinitionId": req_remove_resource['id']
            }

            res_remove_resources.append(res_remove_resource)

        return res_remove_resources

    @staticmethod
    def _make_update_resources(req_update_resources):
        res_update_resources = []
        for req_update_resource in req_update_resources:
            res_update_resource = {
                "resourceDefinitionId": req_update_resource['id']
            }
            res_update_resources.append(res_update_resource)

        return res_update_resources

    @staticmethod
    def _make_vim_assets(image_id, flavour_id="1"):
        # set m1.tiny="1" for flavour_id
        vim_assets = {
            "computeResourceFlavours": [
                {
                    "vimConnectionId": uuidsentinel.vim_connection_id,
                    "vnfdVirtualComputeDescId": "VDU1",
                    "vimFlavourId": flavour_id
                },
                {
                    "vimConnectionId": uuidsentinel.vim_connection_id,
                    "vnfdVirtualComputeDescId": "VDU2",
                    "vimFlavourId": flavour_id
                }
            ],
            "softwareImages": [
                {
                    "vimConnectionId": uuidsentinel.vim_connection_id,
                    "vnfdSoftwareImageId": "VDU1",
                    "vimSoftwareImageId": image_id
                },
                {
                    "vimConnectionId": uuidsentinel.vim_connection_id,
                    "vnfdSoftwareImageId": "VDU2",
                    "vimSoftwareImageId": image_id
                }
            ]
        }

        return vim_assets

    @staticmethod
    def _make_response_template(request_body):
        res = {
            "id": uuidsentinel.__getattr__(request_body['vnfLcmOpOccId']),
            "vnfInstanceId": request_body['vnfInstanceId'],
            "vnfLcmOpOccId": request_body['vnfLcmOpOccId'],
        }
        res["_links"] = {
            "self": {
                # set fake server port.
                "href": os.path.join(
                    'http://localhost:9990',
                    Grant.GRANT_REQ_PATH)},
            "vnfLcmOpOcc": {
                "href": os.path.join(
                    'http://localhost:9890/vnflcm/v1/vnf_lcm_op_occs',
                    request_body['vnfLcmOpOccId'])},
            "vnfInstance": {
                "href": os.path.join(
                    'http://localhost:9890/vnflcm/v1/vnf_instances',
                    request_body['vnfInstanceId'])}}

        return res

    @staticmethod
    def _convert_body_to_dict(body):
        if isinstance(body, str):
            return jsonutils.loads(body)

        return body

    @staticmethod
    def make_inst_response_body(request_body, tenant_id, image_id):
        request_body = Grant._convert_body_to_dict(request_body)
        res = Grant._make_response_template(request_body)
        res["vimConnections"] = Grant._make_vim_connection_info(tenant_id)
        res["zones"] = Grant.ZONES
        if 'addResources' in request_body.keys():
            res["addResources"] = Grant._make_add_resources(
                request_body['addResources'])
        res["vimAssets"] = Grant._make_vim_assets(
            image_id)
        res["additionalParams"] = Grant.ADDITIONAL_PARAMS

        return res

    @staticmethod
    def make_heal_response_body(request_body, tenant_id, image_id):
        request_body = Grant._convert_body_to_dict(request_body)
        res = Grant._make_response_template(request_body)
        res["vimConnections"] = Grant._make_vim_connection_info(tenant_id)
        res["zones"] = Grant.ZONES
        if 'addResources' in request_body.keys():
            res["addResources"] = Grant._make_add_resources(
                request_body['addResources'])
        if 'removeResources' in request_body.keys():
            res["removeResources"] = Grant._make_remove_resources(
                request_body['removeResources'])
        res["vimAssets"] = Grant._make_vim_assets(image_id)

        return res

    @staticmethod
    def make_scaleout_response_body(request_body, tenant_id, image_id):
        request_body = Grant._convert_body_to_dict(request_body)
        res = Grant._make_response_template(request_body)
        res["vimConnections"] = Grant._make_vim_connection_info(tenant_id)
        res["zones"] = Grant.ZONES
        if 'addResources' in request_body.keys():
            res["addResources"] = Grant._make_add_resources(
                request_body['addResources'])
        res["vimAssets"] = Grant._make_vim_assets(
            image_id)

        return res

    @staticmethod
    def make_scalein_response_body(request_body):
        request_body = Grant._convert_body_to_dict(request_body)
        res = Grant._make_response_template(request_body)
        if 'removeResources' in request_body.keys():
            res["removeResources"] = Grant._make_remove_resources(
                request_body['removeResources'])

        return res

    @staticmethod
    def make_term_response_body(request_body):
        request_body = Grant._convert_body_to_dict(request_body)
        res = Grant._make_response_template(request_body)
        if 'removeResources' in request_body.keys():
            res["removeResources"] = Grant._make_remove_resources(
                request_body['removeResources'])

        return res

    @staticmethod
    def make_change_ext_conn_response_body(request_body, tenant_id, image_id):
        request_body = Grant._convert_body_to_dict(request_body)
        res = Grant._make_response_template(request_body)
        res["vimConnections"] = Grant._make_vim_connection_info(tenant_id)
        res["zones"] = Grant.ZONES
        if 'updateResources' in request_body.keys():
            res["updateResources"] = Grant._make_update_resources(
                request_body['updateResources'])
        res["vimAssets"] = Grant._make_vim_assets(image_id)
        res["additionalParams"] = Grant.ADDITIONAL_PARAMS

        return res
